#include <errno.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef RELEASE
#define PRT(format, ...) printf(format, ##__VA_ARGS__);
#else
#define PRT(...) (void)0;
#endif

int *link; ///< suffix link of state[i]
int *len; ///< length of the longest word in state[i]
int *transitions; ///< index in [0, 2 * count - 1]

int sink; ///< last added node
int wordLen; ///< length of the input word        
int count; ///< count of nodes

void init(const char*);
void build(const char*);
void append(char);
int checkTransition(int, char);
int split(int, int, char);
void freeMem();
void printShortestNonAppearingString(const char*);
char *snas();

void init(const char *str) {
  PRT("init()\n");
  
  wordLen = strlen(str);
		
  // initialize arrays
  link = (int*)malloc(2 * wordLen * sizeof(int)); // auxiliary array tracks suffix links
  memset(link, 0x0, 2 * wordLen * sizeof(int));

  len =  (int*)malloc(2 * wordLen * sizeof(int)); // helper to track node's representative's length
  memset(len, 0x0, 2 * wordLen * sizeof(int));

  transitions = (int*)malloc(4 * wordLen * sizeof(int));
  memset(transitions, 0x0, 4 * wordLen * sizeof(int)); // 0 will be invalid transition
  
  // initial values
  link[0] = -1;
  len[0] = 0;

  sink = 0;
  count = 1;
}

void build(const char *word) {
  assert(word != NULL);
  PRT("build()\n");
  
  while (*word) {
    append(*word);
    word++;
  }
}

// Using the marvellous CP-Algorithms Suffix Automaton article
// and the paper at  https://cs.nyu.edu/~mohri/pub/nfac.pdf
// which is based on Blumer & Blumer et. al. (1985).
void append(char c) {
  PRT("append(%c)\n", c);
  
  int newstate = count++;
                
  len[newstate] = len[sink] + 1;
  int parent = sink;
  while (parent != -1 && (checkTransition(parent, c) == 0)) {
    transitions[2 * parent + (c - '0')] = newstate;
    parent = link[parent];
  }

  if (parent == -1) {
    link[newstate] = 0;
  } else {
    int child = checkTransition(parent, c);
    if (len[parent] + 1 == len[child]) {
      link[newstate] = child;
    } else {
      int clone = split(parent, child, c);
      link[child] = clone;
      link[newstate] = clone;
    }
  }
  sink = newstate;
}

int checkTransition(int state, char c) {
  return transitions[2 * state + (c - '0')];
}

int split(int parent, int child, char c) {
  PRT("split(%d, %d, %c)\n", parent, child, c);
  int clone = count++;
  
  len[clone] = len[parent] + 1;
  link[clone] = link[child];
  transitions[2 * clone + 0] = transitions[2 * child + 0];
  transitions[2 * clone + 1] = transitions[2 * child + 1];
  
  while (parent != -1 && checkTransition(parent, c) == child) {
    transitions[2 * parent + (c - '0')] = clone;
    parent = link[parent];
  }

  return clone;
}

void freeMem() {
  if (link) free(link);
  if (len) free(len);
  if (transitions) free(transitions);
}

void printShortestNonAppearingString(const char *word) {
  init(word);
  build(word);

  for (int i = 0; i < count; ++i) {
    PRT("Node %d{%d, %d}: tr[0] = %d, tr[1] = %d\n", i, len[i], link[i], transitions[2 * i],
           transitions[2 * i + 1]);
  }
  
  char *ans = snas();
  printf("Shortest non-appearing string in %s is %s.\n", word, ans);

  free(ans);
  freeMem();
}

typedef struct bfs_t { // bfs node
  int current : 32;
  int prev : 31;
  unsigned char label : 1;
} bfs;

char *snas() {
  PRT("snas()\n");
  
  bfs *queue = (bfs*) malloc(2 * wordLen * sizeof(bfs));
  int head = 0;
  int cnt = 1;
  queue[0].prev = -1;
  
  char *ans = (char*) malloc(wordLen + 1);
  int ansLen = 0;
  
  bfs bt; // helper node for backtracing
  while (1) {
    int curr = queue[head].current;

    int tr;
    if ((tr = transitions[2 * curr + 0]) != 0) {
      queue[cnt].current = tr;
      queue[cnt].prev = head;
      queue[cnt].label = 0;
      ++cnt;
    } else {
      bt.prev = head;
      bt.label = 0;
      break;
    }

    if ((tr = transitions[2 * curr + 1]) != 0) {
      queue[cnt].current = tr;
      queue[cnt].prev = head;
      queue[cnt].label = 1;
      ++cnt;
    } else {
      bt.prev = head;
      bt.label = 1;
      break;
    }

    ++head;
  }

  // backtrack the resulting word
  while (bt.prev != -1) {
    ans[ansLen++] = '0' + abs(bt.label);
    bt.label = queue[bt.prev].label;
    bt.prev = queue[bt.prev].prev;
  }

  // reverse the buffer
  char *revAns = (char*) malloc(ansLen + 1);
  revAns[ansLen] = '\0';
  int i = 0;
  while (i < ansLen) {
    revAns[i] = ans[ansLen - i - 1];
    ++i;
  }

  free(queue);
  free(ans);
  return revAns;
}

int main(int argc, char **argv) {
  if (argc != 2) {
    printf("Invalid number of arguments!\nUsage %s [word]\n", argv[0]);
    exit(2);
  }
  
  const char *word = argv[1];
  
  if (word == NULL) {
    exit(1);
  }

  printShortestNonAppearingString(word);
}
